# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from tensorflow_model_optimization.python.core.quantization.keras.default_8bit.default_8bit_quantize_configs import \
    NoOpQuantizeConfig
from tensorflow_model_optimization.python.core.quantization.keras.quantize_config import QuantizeConfig

from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.keras.default_framework_info import DEFAULT_KERAS_INFO
from model_compression_toolkit.exporter.fully_quantized.keras.builder.quantizer_to_node import \
    get_weights_quantizer_for_node, get_activations_quantizer_for_node
from model_compression_toolkit.exporter.fully_quantized.keras.quantize_configs.activation_quantize_config import \
    ActivationQuantizeConfig
from model_compression_toolkit.exporter.fully_quantized.keras.quantize_configs.weights_activation_quantize_config \
    import \
    WeightsActivationQuantizeConfig
from model_compression_toolkit.exporter.fully_quantized.keras.quantize_configs.weights_quantize_config import \
    WeightsQuantizeConfig


def get_quantization_config(node: BaseNode) -> QuantizeConfig:
    """
    Create a QuantizeConfig to wrap a layer for its corresponding node.

    Args:
        node: Node to create a QuantizeConfig for.

    Returns:
        QuantizeConfig to use for wrapping the layer from the passed node.
    """

    if node.is_weights_quantization_enabled() and not node.is_activation_quantization_enabled():
        weight_attrs = DEFAULT_KERAS_INFO.get_kernel_op_attributes(node.type)
        return WeightsQuantizeConfig(weight_attrs=weight_attrs,
                                     w_quantizer=get_weights_quantizer_for_node(node,
                                                                                weight_attrs))

    elif not node.is_weights_quantization_enabled() and node.is_activation_quantization_enabled():
        return ActivationQuantizeConfig(activation_quantizer=get_activations_quantizer_for_node(node))

    elif not node.is_weights_quantization_enabled() and not node.is_activation_quantization_enabled():
        return NoOpQuantizeConfig()

    weight_attrs = DEFAULT_KERAS_INFO.get_kernel_op_attributes(node.type)
    return WeightsActivationQuantizeConfig(activation_quantizer=get_activations_quantizer_for_node(node),
                                           w_quantizer=get_weights_quantizer_for_node(node,
                                                                                      weight_attrs),
                                           weight_attrs=DEFAULT_KERAS_INFO.get_kernel_op_attributes(node.type))



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
from typing import List

from model_compression_toolkit.core.common import Graph, BaseNode
from model_compression_toolkit.core.common.constants import DEFAULT_CANDIDATE_BITWIDTH
from model_compression_toolkit.core.common.quantization.candidate_node_quantization_config import \
    CandidateNodeQuantizationConfig


def filter_nodes_candidates(graph_to_filter: Graph):
    """
    Filters the graph's nodes candidates configuration list.
    We apply this after mark activation operation to eliminate nodes that their activation are no longer being quantized
    from the mixed-precision search.
    Updating the lists is preformed inplace on the graph object.

    Args:
        graph_to_filter: Graph for which to add quantization info to each node.
    """
    graph = copy.deepcopy(graph_to_filter)
    nodes = list(graph.nodes)
    for n in nodes:
        n.candidates_quantization_cfg = filter_node_candidates(node=n)

    return graph


def filter_node_candidates(node: BaseNode) -> List[CandidateNodeQuantizationConfig]:
    """
    Updates a node's candidates configuration list.
    If the node's weights quantization is disabled (or it only has activations to quantize), then the updated list
    will have a candidate with any of the different original activation bitwidths candidates and a default value
    for its weights bitwidth (that doesn't have any impact on the quantization or the mixed-precision search.
    If the node's activation quantization is disabled, the same filtering applied for the weights bitwidth candidates.

    Args:
        node: Node to set its quantization configurations.
    """

    filtered_candidates = copy.deepcopy(node.candidates_quantization_cfg)

    if not node.is_weights_quantization_enabled() and not node.is_activation_quantization_enabled():
        # If both weights and activation quantization are disabled, but for some reason the node has multiple candidates
        # then replace it with a single dummy candidate with default bit-width values.
        single_dummy_candidate = filtered_candidates[0]
        single_dummy_candidate.activation_quantization_cfg.activation_n_bits = DEFAULT_CANDIDATE_BITWIDTH
        single_dummy_candidate.weights_quantization_cfg.weights_n_bits = DEFAULT_CANDIDATE_BITWIDTH
        filtered_candidates = [single_dummy_candidate]

    elif not node.is_activation_quantization_enabled():
        # Remove candidates that have duplicated weights candidates for node with disabled activation quantization.
        # Replacing the activation n_bits in the remained configurations with default value to prevent confusion.
        seen_candidates = set()
        filtered_candidates = [candidate for candidate in filtered_candidates if
                               candidate.weights_quantization_cfg not in seen_candidates
                               and not seen_candidates.add(candidate.weights_quantization_cfg)]

        for c in filtered_candidates:
            c.activation_quantization_cfg.activation_n_bits = DEFAULT_CANDIDATE_BITWIDTH

    elif not node.is_weights_quantization_enabled():
        # Remove candidates that have duplicated activation candidates for node with disabled weights quantization.
        # Replacing the weights n_bits in the remained configurations with default value to prevent confusion.
        seen_candidates = set()
        filtered_candidates = [candidate for candidate in filtered_candidates if
                               candidate.activation_quantization_cfg not in seen_candidates
                               and not seen_candidates.add(candidate.activation_quantization_cfg)]

        for c in filtered_candidates:
            c.weights_quantization_cfg.weights_n_bits = DEFAULT_CANDIDATE_BITWIDTH

    return filtered_candidates



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy
from enum import Enum

from typing import List

from model_compression_toolkit.core.common import Graph, Logger
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi import KPI
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi_functions_mapping import kpi_functions_mapping
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_search_manager import MixedPrecisionSearchManager
from model_compression_toolkit.core.common.mixed_precision.search_methods.linear_programming import \
    mp_integer_programming_search
from model_compression_toolkit.core.common.framework_info import FrameworkInfo


# When adding a new search_methods MP configuration method, these enum and factory dictionary
# should be updated with it's kind and a search_method implementation.
from model_compression_toolkit.core.common.mixed_precision.sensitivity_evaluation import SensitivityEvaluation


class BitWidthSearchMethod(Enum):
    INTEGER_PROGRAMMING = 0


search_methods = {
    BitWidthSearchMethod.INTEGER_PROGRAMMING: mp_integer_programming_search}


def search_bit_width(graph_to_search_cfg: Graph,
                     fw_info: FrameworkInfo,
                     fw_impl: FrameworkImplementation,
                     target_kpi: KPI,
                     sensitivity_evaluator: SensitivityEvaluation = None,
                     search_method: BitWidthSearchMethod = BitWidthSearchMethod.INTEGER_PROGRAMMING) -> List[int]:
    """
    Search for an MP configuration for a given graph. Given a search_method method (by default, it's linear
    programming), we use the sensitivity_evaluator object that provides a function to compute an
    evaluation for the expected sensitivity for a bit-width configuration.
    Then, and after computing the KPI for each node in the graph for each bit-width in the search space,
    we search for the optimal solution, given some target_kpi, the solution should fit.
    target_kpi have to be passed. If it was not passed, the facade is not supposed to get here by now.

    Args:
        graph_to_search_cfg: Graph to search a MP configuration for.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize).
        fw_impl: FrameworkImplementation object with specific framework methods implementation.
        target_kpi: Target KPI to bound our feasible solution space s.t the configuration does not violate it.
        sensitivity_evaluator: A SensitivityEvaluation which provides a function that evaluates the sensitivity of
            a bit-width configuration for the MP model.
        search_method: BitWidthSearchMethod to define which searching method to use.

    Returns:
        A MP configuration for the graph (list of integers, where the index in the list, is the node's
        index in the graph, when the graph is topology sorted, and the value in this index is the
        bit-width index on the node).

    """

    # target_kpi have to be passed. If it was not passed, the facade is not supposed to get here by now.
    if target_kpi is None:
        Logger.critical('Target KPI have to be passed for search_methods bit-width configuration')

    graph = copy.deepcopy(graph_to_search_cfg)  # Copy graph before searching

    # Each pair of (KPI method, KPI aggregation) should match to a specific provided kpi target
    kpi_functions = kpi_functions_mapping

    # Instantiate a manager object
    search_manager = MixedPrecisionSearchManager(graph,
                                                 fw_info,
                                                 fw_impl,
                                                 sensitivity_evaluator,
                                                 kpi_functions)

    if search_method in search_methods:  # Get a specific search function
        search_method_fn = search_methods.get(search_method)
    else:
        raise NotImplemented

    # Search for the desired mixed-precision configuration
    result_bit_cfg = search_method_fn(search_manager,
                                      target_kpi)

    return result_bit_cfg



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from collections import Callable

from model_compression_toolkit.core.common.target_platform import QuantizationMethod
from model_compression_toolkit.core.common.quantization.quantization_params_generation.kmeans_params import kmeans_tensor
from model_compression_toolkit.core.common.quantization.quantization_params_generation.lut_kmeans_params import \
    lut_kmeans_tensor, lut_kmeans_histogram
from model_compression_toolkit.core.common.quantization.quantization_params_generation.symmetric_selection import \
    symmetric_selection_tensor, symmetric_selection_histogram
from model_compression_toolkit.core.common.quantization.quantization_params_generation.uniform_selection import \
    uniform_selection_histogram, uniform_selection_tensor
from model_compression_toolkit.core.common.quantization.quantization_params_generation.power_of_two_selection import \
    power_of_two_selection_tensor, power_of_two_selection_histogram

def get_activation_quantization_params_fn(activation_quantization_method: QuantizationMethod) -> Callable:
    """
    Generate a function for finding activation quantization parameters.

    Args:
        activation_quantization_method: Which quantization method to use for activations.
    Returns:
        A function to find the quantization parameters.

    """
    if activation_quantization_method == QuantizationMethod.POWER_OF_TWO:
        params_fn = power_of_two_selection_histogram
    elif activation_quantization_method == QuantizationMethod.SYMMETRIC:
        params_fn = symmetric_selection_histogram
    elif activation_quantization_method == QuantizationMethod.UNIFORM:
        params_fn = uniform_selection_histogram
    elif activation_quantization_method == QuantizationMethod.LUT_QUANTIZER:
        params_fn = lut_kmeans_histogram
    else:
        raise Exception(
            f'No params function for the configuration of quantization method {activation_quantization_method}')
    return params_fn


def get_weights_quantization_params_fn(weights_quantization_method: QuantizationMethod) -> Callable:
    """
    Generate a function for finding weights quantization parameters.

    Args:
        weights_quantization_method: Which quantization method to use for weights.
    Returns:
        A function to find the quantization parameters.

    """
    if weights_quantization_method == QuantizationMethod.POWER_OF_TWO:
        params_fn = power_of_two_selection_tensor
    elif weights_quantization_method == QuantizationMethod.SYMMETRIC:
        params_fn = symmetric_selection_tensor
    elif weights_quantization_method == QuantizationMethod.UNIFORM:
        params_fn = uniform_selection_tensor
    elif weights_quantization_method == QuantizationMethod.KMEANS:
        params_fn = kmeans_tensor
    elif weights_quantization_method == QuantizationMethod.LUT_QUANTIZER:
        params_fn = lut_kmeans_tensor
    else:
        raise Exception(
            f'No params function for the configuration of quantization method {weights_quantization_method}')
    return params_fn



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi import KPITarget
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi_aggregation_methods import MpKpiAggregation
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi_methods import MpKpiMetric


# When adding a KPITarget that we want to consider in our mp search,
# a matching pair of kpi_tools computation function and a kpi_tools
# aggregation function should be added to this dictionary
kpi_functions_mapping = {KPITarget.WEIGHTS: (MpKpiMetric.WEIGHTS_SIZE, MpKpiAggregation.SUM),
                         KPITarget.ACTIVATION: (MpKpiMetric.ACTIVATION_OUTPUT_SIZE, MpKpiAggregation.MAX),
                         KPITarget.TOTAL: (MpKpiMetric.TOTAL_WEIGHTS_ACTIVATION_SIZE, MpKpiAggregation.TOTAL)}



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf
import itertools
from tensorflow.keras.layers import Dense, DepthwiseConv2D, Conv2D, Conv2DTranspose
from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core.common import BaseNode, Graph, BaseSubstitution
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core.common.graph.virtual_activation_weights_node import VirtualSplitWeightsNode, \
    VirtualSplitActivationNode


class WeightsActivationSplit(BaseSubstitution):
    def __init__(self):
        """
        Matches: (DepthwiseConv2D, Conv2D, Dense, Conv2DTranspose)
        """
        op2d_node = NodeOperationMatcher(DepthwiseConv2D) | \
                    NodeOperationMatcher(Conv2D) | \
                    NodeOperationMatcher(Dense) | \
                    NodeOperationMatcher(Conv2DTranspose)

        super().__init__(matcher_instance=op2d_node)

    def substitute(self,
                   graph: Graph,
                   node: BaseNode) -> Graph:
        """
        Decompose a linear node into two nodes - one with the linear operations (a weights node) and one with
        the activation operation (with an identity function that just passes the node's output, but allows to
        quantize it according to the node's activation quantization configuration candidates).
        The two new virtual nodes are connected with an edge [weights node --> activation node].
        Note that the node is split only if its candidates list is composite, that is, it contains all the combinations
        of activation and weights bit-width that exists in any of its candidates.

        Args:
            graph: Graph we apply the substitution on.
            node: Node to split.

        Returns:
            Graph after applying the substitution.
        """

        if not node.is_all_weights_candidates_equal() and not node.is_all_activation_candidates_equal():
            # Node has both different weights and different activation configuration candidates
            weights_bits = [c.weights_quantization_cfg.weights_n_bits for c in node.get_unique_weights_candidates()]
            activation_bits = [c.activation_quantization_cfg.activation_n_bits for c in node.get_unique_activation_candidates()]
            expected_candidates = list(itertools.product(weights_bits, activation_bits))
            all_candidates_bits = [(c.weights_quantization_cfg.weights_n_bits,
                                    c.activation_quantization_cfg.activation_n_bits) for c in node.candidates_quantization_cfg]
            if not set(expected_candidates).issubset(all_candidates_bits):
                # Node is not composite, therefore, can't be split
                Logger.critical(f"The graph contains a node {node.name} with non composite candidates."
                                f"In order to run mixed-precision search with BOPS target KPI, "
                                f"all model layers should be composite.")

        weights_node = VirtualSplitWeightsNode(node)
        activation_node = VirtualSplitActivationNode(node, tf.keras.layers.Activation)

        # Update graph
        graph.add_node(weights_node)
        graph.add_node(activation_node)
        graph.reconnect_in_edges(current_node=node, new_node=weights_node)
        graph.reconnect_out_edges(current_node=node, new_node=activation_node)
        graph.replace_output_node(current_node=node, new_node=activation_node)
        graph.add_edge(weights_node,
                       activation_node,
                       source_index=0,
                       sink_index=0)
        graph.remove_node(node)

        return graph

